import io
import os
import pickle
import sys

import boto3
from dagster_aws.s3.file_manager import S3FileHandle, S3FileManager

from dagster.core.execution.plan.external_step import PICKLED_EVENTS_FILE_NAME, run_step_from_ref


def main(step_run_ref_bucket, s3_dir_key):
    session = boto3.client('s3')
    file_manager = S3FileManager(session, step_run_ref_bucket, '')
    file_handle = S3FileHandle(step_run_ref_bucket, s3_dir_key)
    step_run_ref_data = file_manager.read_data(file_handle)

    step_run_ref = pickle.loads(step_run_ref_data)

    events = list(run_step_from_ref(step_run_ref))
    file_obj = io.BytesIO(pickle.dumps(events))
    events_key = os.path.dirname(s3_dir_key) + '/' + PICKLED_EVENTS_FILE_NAME
    session.put_object(Body=file_obj, Bucket=step_run_ref_bucket, Key=events_key)


if __name__ == '__main__':
    main(sys.argv[1], sys.argv[2])



import base64
import getpass
import os

import boto3
import dagster_aws
import pytest
import six
from dagster_aws.cli.aws_util import (
    DEFAULT_AMI,
    DEFAULT_INSTANCE_TYPE,
    VPC_CREATION_WARNING,
    create_ec2_instance,
    create_key_pair,
    create_rds_instance,
    create_security_group,
    get_validated_ami_id,
    select_region,
    select_vpc,
)
from dagster_aws.cli.config import EC2Config, RDSConfig
from moto import mock_ec2, mock_rds, mock_secretsmanager
from six import StringIO


@mock_ec2
def test_select_region(monkeypatch, capsys):
    region = 'us-east-1'
    monkeypatch.setattr(
        'sys.stdin', StringIO('does-not-exist\n{region}\n'.format(region=region)),
    )
    res = select_region(None)
    assert res == region

    captured = capsys.readouterr()
    assert 'Error: invalid choice: does-not-exist' in captured.out


@mock_ec2
def test_select_region_prev_config(capsys):
    region = 'us-east-1'
    prev_config = EC2Config(region=region)
    assert select_region(prev_config) == region
    captured = capsys.readouterr()
    assert 'Found existing region, continuing with {region}'.format(region=region) in captured.out


@mock_ec2
def test_select_vpc(monkeypatch, capsys):
    client = boto3.client('ec2', region_name='us-east-1')
    ec2 = boto3.resource('ec2', region_name='us-east-1')

    filters = [{'Name': 'isDefault', 'Values': ['true']}]
    default_vpcs = list(ec2.vpcs.filter(Filters=filters))  # pylint: disable=no-member

    default_vpc = default_vpcs[0].id
    monkeypatch.setattr(
        'sys.stdin', StringIO('does-not-exist\n{default_vpc}\n'.format(default_vpc=default_vpc)),
    )
    vpc = select_vpc(client, ec2)

    captured = capsys.readouterr()

    assert vpc == default_vpc
    assert VPC_CREATION_WARNING in captured.out
    assert 'Select an existing VPC ID to use' in captured.out
    assert 'Specified VPC does-not-exist does not exist' in captured.out

    # Test missing default VPCs
    for vpc in default_vpcs:
        client.delete_vpc(VpcId=vpc.id)

    with pytest.raises(SystemExit) as exc_info:
        select_vpc(client, ec2)

    captured = capsys.readouterr()

    assert exc_info.value.code == 1
    assert 'AWS account does not have a default VPC' in captured.err


@mock_ec2
def test_select_vpc_created(monkeypatch):
    client = boto3.client('ec2', region_name='us-east-1')
    ec2 = boto3.resource('ec2', region_name='us-east-1')

    vpc = ec2.create_vpc(CidrBlock='172.16.0.0/16')  # pylint: disable=no-member
    monkeypatch.setattr(
        'sys.stdin', StringIO('{vpc_id}\n'.format(vpc_id=vpc.id)),
    )
    selected_vpc = select_vpc(client, ec2)
    assert vpc.id == selected_vpc


@mock_ec2
def test_create_security_groups(monkeypatch, capsys):
    client = boto3.client('ec2', region_name='us-east-1')
    ec2 = boto3.resource('ec2', region_name='us-east-1')

    filters = [{'Name': 'isDefault', 'Values': ['true']}]
    default_vpc = list(ec2.vpcs.filter(Filters=filters))[0].id  # pylint: disable=no-member

    print(
        "client.describe_security_groups()['SecurityGroups']",
        client.describe_security_groups()['SecurityGroups'],
    )

    monkeypatch.setattr(
        'sys.stdin', StringIO('default\nmynewcoolgroup\n'),
    )
    prev_config = None
    res = create_security_group(prev_config, client, ec2, default_vpc)

    filters = [{'Name': 'isDefault', 'Values': ['true']}]
    sg_described = client.describe_security_groups(GroupIds=[res])['SecurityGroups'][0]

    assert sg_described['Description'] == 'dagit Security Group'
    assert sg_described['GroupName'] == 'mynewcoolgroup'

    assert sg_described['IpPermissions'] == [
        {
            'FromPort': 3000,
            'IpProtocol': 'tcp',
            'IpRanges': [{'CidrIp': '0.0.0.0/0'}],
            'ToPort': 3000,
            'UserIdGroupPairs': [],
        },
        {
            'FromPort': 22,
            'IpProtocol': 'tcp',
            'IpRanges': [{'CidrIp': '0.0.0.0/0'}],
            'ToPort': 22,
            'UserIdGroupPairs': [],
        },
    ]
    assert sg_described['Tags'] == [{'Key': 'Name', 'Value': 'mynewcoolgroup'}]

    next_group = 'myothercoolgroup'
    prev_config = EC2Config(security_group_id=next_group)
    assert create_security_group(prev_config, client, ec2, default_vpc) == next_group
    captured = capsys.readouterr()
    assert (
        'Found existing security group, continuing with {group}'.format(group=next_group)
        in captured.out
    )


@mock_ec2
def test_create_key_pair(monkeypatch, capsys, tmp_path):
    tmp_path = str(tmp_path)

    client = boto3.client('ec2', region_name='us-east-1')

    existing = 'foobar-keypair'
    client.create_key_pair(KeyName=existing)

    new_keypair = 'my-new-keypair'
    monkeypatch.setattr(
        'sys.stdin',
        StringIO('{existing}\n{new_keypair}\n'.format(existing=existing, new_keypair=new_keypair)),
    )

    prev_config = None
    res = create_key_pair(prev_config, client, tmp_path)
    assert res[0] == new_keypair
    assert res[1] == os.path.join(tmp_path, 'keys/%s.pem' % new_keypair)

    captured = capsys.readouterr()
    assert (
        'Specified key pair {existing} already exists, won\'t create'.format(existing=existing)
        in captured.out
    )

    # test existing
    key_file_path = os.path.join(tmp_path, 'mykey')
    open(key_file_path, 'a').close()
    prev_config = EC2Config(key_pair_name='mykey', key_file_path=key_file_path)
    res = create_key_pair(prev_config, client, tmp_path)
    assert res == ('mykey', key_file_path)

    captured = capsys.readouterr()
    assert 'Found existing key pair mykey' in captured.out


@mock_ec2
def test_get_validated_ami_id(monkeypatch, capsys):
    client = boto3.client('ec2', region_name='us-east-1')

    source_ami_id = 'ami-03cf127a'

    monkeypatch.setattr(
        'sys.stdin',
        StringIO('does-not-exist\n{source_ami_id}\n'.format(source_ami_id=source_ami_id)),
    )
    res = get_validated_ami_id(client)

    assert res == source_ami_id
    captured = capsys.readouterr()

    prompt = (
        'Choose an AMI to use (must be Debian-based) [default is %s (For us-west-1 region only)]:'
        % DEFAULT_AMI
    )
    assert captured.out.strip() == u'{prompt} \u274C  Specified AMI does not exist in the chosen region, fix to continue\n\n{prompt}'.format(
        prompt=prompt
    )


@mock_ec2
def test_create_ec2_instance(monkeypatch):
    def _validate_user_data(ec2_instance, script_name):
        # Validate user data is init script
        with open(
            os.path.join(os.path.dirname(dagster_aws.cli.aws_util.__file__), 'shell', script_name),
            'rb',
        ) as f:
            init_script = six.ensure_str(f.read())

        user_data_b64 = ec2_instance.describe_attribute(Attribute='userData')['UserData']['Value']
        user_data_str = six.ensure_str(base64.b64decode(user_data_b64))

        assert user_data_str == init_script

    client = boto3.client('ec2', region_name='us-west-1')
    ec2 = boto3.resource('ec2', region_name='us-west-1')

    filters = [{'Name': 'isDefault', 'Values': ['true']}]
    default_vpc = list(ec2.vpcs.filter(Filters=filters))[0].id  # pylint: disable=no-member

    keypair_name = 'foobar-keypair'
    client.create_key_pair(KeyName=keypair_name)

    # Test init-stable.sh
    instance_type = 't3.large'
    monkeypatch.setattr(
        'sys.stdin',
        StringIO('\n{instance_type}\nmycoolinstance\n'.format(instance_type=instance_type)),
    )
    prev_config = None
    sg_id = create_security_group(prev_config, client, ec2, default_vpc)

    ec2_instance = create_ec2_instance(
        client, ec2, sg_id, DEFAULT_AMI, keypair_name, use_master=False
    )

    assert ec2_instance.instance_type == instance_type
    assert ec2_instance.image_id == DEFAULT_AMI
    assert ec2_instance.tags == [{'Key': 'Name', 'Value': 'mycoolinstance'}]
    assert sg_id in [g['GroupId'] for g in ec2_instance.security_groups]
    _validate_user_data(ec2_instance, 'init-stable.sh')

    # Test init.sh
    monkeypatch.setattr(
        'sys.stdin', StringIO('\n\n\n'),
    )
    prev_config = None

    ec2_instance = create_ec2_instance(
        client, ec2, sg_id, DEFAULT_AMI, keypair_name, use_master=True
    )

    assert ec2_instance.instance_type == DEFAULT_INSTANCE_TYPE
    assert ec2_instance.image_id == DEFAULT_AMI
    assert sg_id in [g['GroupId'] for g in ec2_instance.security_groups]
    assert ec2_instance.tags == [{'Key': 'Name', 'Value': 'dagit-ec2-%s' % getpass.getuser()}]
    _validate_user_data(ec2_instance, 'init.sh')


@mock_rds
@mock_secretsmanager
def test_create_rds_instance(monkeypatch, tmp_path):
    tmp_path = str(tmp_path)

    monkeypatch.setattr(
        'sys.stdin', StringIO('\n'),
    )
    region = 'us-east-1'
    rds = boto3.client('rds', region_name=region)

    # Test don't create
    rds_instance = create_rds_instance(tmp_path, region)
    assert rds_instance is None

    # Test create
    instance_type = 'db.t3.2xlarge'
    instance_name = 'foobardb'
    monkeypatch.setattr(
        'sys.stdin',
        StringIO(
            'Y\n{instance_type}\n{instance_name}\n'.format(
                instance_type=instance_type, instance_name=instance_name
            )
        ),
    )
    rds_instance = create_rds_instance(tmp_path, 'us-east-1')
    db_instance = rds.describe_db_instances(DBInstanceIdentifier=instance_name)['DBInstances'][0]

    assert rds_instance.instance_name == instance_name
    assert rds_instance.instance_type == instance_type
    assert rds_instance.storage_size_gb == 20
    assert rds_instance.db_engine == 'postgres'
    assert rds_instance.db_name == 'dagster'
    assert rds_instance.username == 'dagster'

    assert db_instance['DBInstanceIdentifier'] == instance_name
    assert db_instance['DBInstanceClass'] == instance_type
    assert db_instance['Engine'] == 'postgres'
    assert db_instance['DBInstanceStatus'] == 'available'
    assert db_instance['MasterUsername'] == 'dagster'
    assert db_instance['DBName'] == 'dagster'
    assert db_instance['Endpoint']['Port'] == 5432

    assert db_instance['Endpoint']['Address'] == rds_instance.instance_uri
    assert db_instance['AllocatedStorage'] == rds_instance.storage_size_gb
    assert db_instance['EngineVersion'] == rds_instance.db_engine_version


def test_create_rds_instance_already_exists(capsys, tmp_path):

    rds_config = RDSConfig(
        instance_name='foo', instance_uri='foo-bar.amazonaws.com', password='baz'
    )
    rds_config.save(str(tmp_path))
    prev_config = create_rds_instance(str(tmp_path), 'us-west-1')
    assert prev_config == rds_config

    captured = capsys.readouterr()
    assert 'Found existing RDS database, continuing with foo-bar.amazonaws.com' in captured.out



import os
from contextlib import contextmanager

from dagster import Field, check, seven
from dagster.core.storage.compute_log_manager import (
    MAX_BYTES_FILE_READ,
    ComputeIOType,
    ComputeLogFileData,
    ComputeLogManager,
)
from dagster.core.storage.local_compute_log_manager import IO_TYPE_EXTENSION, LocalComputeLogManager
from dagster.serdes import ConfigurableClass, ConfigurableClassData
from dagster.utils import ensure_dir, ensure_file

from .utils import create_s3_session


class S3ComputeLogManager(ComputeLogManager, ConfigurableClass):
    '''Logs solid compute function stdout and stderr to S3.

    Users should not instantiate this class directly. Instead, use a YAML block in ``dagster.yaml``
    such as the following:

    .. code-block:: YAML

        compute_log_manager:
          module: dagster_aws.s3.compute_log_manager
          class: S3ComputeLogManager
          config:
            bucket: "mycorp-dagster-compute-logs"
            local_dir: "/tmp/cool"
            prefix: "dagster-test-"

    Args:
        bucket (str): The name of the s3 bucket to which to log.
        local_dir (Optional[str]): Path to the local directory in which to stage logs. Default:
            ``dagster.seven.get_system_temp_directory()``.
        prefix (Optional[str]): Prefix for the log file keys.
        inst_data (Optional[ConfigurableClassData]): Serializable representation of the compute
            log manager when newed up from config.
    '''

    def __init__(self, bucket, local_dir=None, inst_data=None, prefix='dagster'):
        self._s3_session = create_s3_session()
        self._s3_bucket = check.str_param(bucket, 'bucket')
        self._s3_prefix = check.str_param(prefix, 'prefix')
        self._download_urls = {}

        # proxy calls to local compute log manager (for subscriptions, etc)
        if not local_dir:
            local_dir = seven.get_system_temp_directory()

        self.local_manager = LocalComputeLogManager(local_dir)
        self._inst_data = check.opt_inst_param(inst_data, 'inst_data', ConfigurableClassData)

    @contextmanager
    def _watch_logs(self, pipeline_run, step_key=None):
        # proxy watching to the local compute log manager, interacting with the filesystem
        with self.local_manager._watch_logs(  # pylint: disable=protected-access
            pipeline_run, step_key
        ):
            yield

    @property
    def inst_data(self):
        return self._inst_data

    @classmethod
    def config_type(cls):
        return {
            'bucket': str,
            'local_dir': Field(str, is_required=False),
            'prefix': Field(str, is_required=False, default_value='dagster'),
        }

    @staticmethod
    def from_config_value(inst_data, config_value):
        return S3ComputeLogManager(inst_data=inst_data, **config_value)

    def get_local_path(self, run_id, key, io_type):
        return self.local_manager.get_local_path(run_id, key, io_type)

    def on_watch_start(self, pipeline_run, step_key):
        self.local_manager.on_watch_start(pipeline_run, step_key)

    def on_watch_finish(self, pipeline_run, step_key):
        self.local_manager.on_watch_finish(pipeline_run, step_key)
        key = self.local_manager.get_key(pipeline_run, step_key)
        self._upload_from_local(pipeline_run.run_id, key, ComputeIOType.STDOUT)
        self._upload_from_local(pipeline_run.run_id, key, ComputeIOType.STDERR)

    def is_watch_completed(self, run_id, key):
        return self.local_manager.is_watch_completed(run_id, key)

    def download_url(self, run_id, key, io_type):
        if not self.is_watch_completed(run_id, key):
            return self.local_manager.download_url(run_id, key, io_type)
        key = self._bucket_key(run_id, key, io_type)
        if key in self._download_urls:
            return self._download_urls[key]
        url = self._s3_session.generate_presigned_url(
            ClientMethod='get_object', Params={'Bucket': self._s3_bucket, 'Key': key}
        )
        self._download_urls[key] = url
        return url

    def read_logs_file(self, run_id, key, io_type, cursor=0, max_bytes=MAX_BYTES_FILE_READ):
        if self._should_download(run_id, key, io_type):
            self._download_to_local(run_id, key, io_type)
        data = self.local_manager.read_logs_file(run_id, key, io_type, cursor, max_bytes)
        return self._from_local_file_data(run_id, key, io_type, data)

    def on_subscribe(self, subscription):
        self.local_manager.on_subscribe(subscription)

    def _should_download(self, run_id, key, io_type):
        local_path = self.get_local_path(run_id, key, io_type)
        if os.path.exists(local_path):
            return False
        s3_objects = self._s3_session.list_objects(
            Bucket=self._s3_bucket, Prefix=self._bucket_key(run_id, key, io_type)
        )
        return len(s3_objects) > 0

    def _from_local_file_data(self, run_id, key, io_type, local_file_data):
        is_complete = self.is_watch_completed(run_id, key)
        path = (
            's3://{}/{}'.format(self._s3_bucket, self._bucket_key(run_id, key, io_type))
            if is_complete
            else local_file_data.path
        )

        return ComputeLogFileData(
            path,
            local_file_data.data,
            local_file_data.cursor,
            local_file_data.size,
            self.download_url(run_id, key, io_type),
        )

    def _upload_from_local(self, run_id, key, io_type):
        path = self.get_local_path(run_id, key, io_type)
        ensure_file(path)
        key = self._bucket_key(run_id, key, io_type)
        with open(path, 'rb') as data:
            self._s3_session.upload_fileobj(data, self._s3_bucket, key)

    def _download_to_local(self, run_id, key, io_type):
        path = self.get_local_path(run_id, key, io_type)
        ensure_dir(os.path.dirname(path))
        with open(path, 'wb') as fileobj:
            self._s3_session.download_fileobj(
                self._s3_bucket, self._bucket_key(run_id, key, io_type), fileobj
            )

    def _bucket_key(self, run_id, key, io_type):
        check.inst_param(io_type, 'io_type', ComputeIOType)
        extension = IO_TYPE_EXTENSION[io_type]
        paths = [
            self._s3_prefix,
            'storage',
            run_id,
            'compute_logs',
            '{}.{}'.format(key, extension),
        ]
        return '/'.join(paths)  # s3 path delimiter



import logging
import sys
from io import BytesIO, StringIO

import boto3

from dagster import check
from dagster.core.definitions.events import ObjectStoreOperation, ObjectStoreOperationType
from dagster.core.storage.object_store import ObjectStore
from dagster.core.types.marshal import SerializationStrategy


class S3ObjectStore(ObjectStore):
    def __init__(self, bucket, s3_session=None):
        self.bucket = check.str_param(bucket, 'bucket')
        self.s3 = s3_session or boto3.client('s3')
        self.s3.head_bucket(Bucket=bucket)
        super(S3ObjectStore, self).__init__('s3', sep='/')

    def set_object(self, key, obj, serialization_strategy=None):
        check.str_param(key, 'key')

        logging.info('Writing S3 object at: ' + self.uri_for_key(key))

        # cannot check obj since could be arbitrary Python object
        check.inst_param(
            serialization_strategy, 'serialization_strategy', SerializationStrategy
        )  # cannot be none here

        if self.has_object(key):
            logging.warning('Removing existing S3 key: {key}'.format(key=key))
            self.rm_object(key)

        with BytesIO() as bytes_io:
            if serialization_strategy.write_mode == 'w' and sys.version_info >= (3, 0):
                with StringIO() as string_io:
                    string_io = StringIO()
                    serialization_strategy.serialize(obj, string_io)
                    string_io.seek(0)
                    bytes_io.write(string_io.read().encode('utf-8'))
            else:
                serialization_strategy.serialize(obj, bytes_io)
            bytes_io.seek(0)
            self.s3.put_object(Bucket=self.bucket, Key=key, Body=bytes_io)

        return ObjectStoreOperation(
            op=ObjectStoreOperationType.SET_OBJECT,
            key=self.uri_for_key(key),
            dest_key=None,
            obj=obj,
            serialization_strategy_name=serialization_strategy.name,
            object_store_name=self.name,
        )

    def get_object(self, key, serialization_strategy=None):
        check.str_param(key, 'key')
        check.param_invariant(len(key) > 0, 'key')
        check.inst_param(
            serialization_strategy, 'serialization_strategy', SerializationStrategy
        )  # cannot be none here

        # FIXME we need better error handling for object store
        obj = serialization_strategy.deserialize(
            BytesIO(self.s3.get_object(Bucket=self.bucket, Key=key)['Body'].read())
            if serialization_strategy.read_mode == 'rb'
            else StringIO(
                self.s3.get_object(Bucket=self.bucket, Key=key)['Body']
                .read()
                .decode(serialization_strategy.encoding)
            )
        )
        return ObjectStoreOperation(
            op=ObjectStoreOperationType.GET_OBJECT,
            key=self.uri_for_key(key),
            dest_key=None,
            obj=obj,
            serialization_strategy_name=serialization_strategy.name,
            object_store_name=self.name,
        )

    def has_object(self, key):
        check.str_param(key, 'key')
        check.param_invariant(len(key) > 0, 'key')

        key_count = self.s3.list_objects_v2(Bucket=self.bucket, Prefix=key)['KeyCount']
        return bool(key_count > 0)

    def rm_object(self, key):
        check.str_param(key, 'key')
        check.param_invariant(len(key) > 0, 'key')

        def delete_for_results(store, results):
            store.s3.delete_objects(
                Bucket=store.bucket,
                Delete={'Objects': [{'Key': result['Key']} for result in results['Contents']]},
            )

        if self.has_object(key):
            results = self.s3.list_objects_v2(Bucket=self.bucket, Prefix=key)
            delete_for_results(self, results)

            continuation = results['IsTruncated']
            while continuation:
                continuation_token = results['NextContinuationToken']
                results = self.s3.list_objects_v2(
                    Bucket=self.bucket, Prefix=key, ContinuationToken=continuation_token
                )
                delete_for_results(self, results)
                continuation = results['IsTruncated']

        return ObjectStoreOperation(
            op=ObjectStoreOperationType.RM_OBJECT,
            key=self.uri_for_key(key),
            dest_key=None,
            obj=None,
            serialization_strategy_name=None,
            object_store_name=self.name,
        )

    def cp_object(self, src, dst):
        check.str_param(src, 'src')
        check.str_param(dst, 'dst')

        self.s3.copy_object(
            Bucket=self.bucket, Key=dst, CopySource={'Bucket': self.bucket, 'Key': src}
        )

        return ObjectStoreOperation(
            op=ObjectStoreOperationType.CP_OBJECT,
            key=self.uri_for_key(src),
            dest_key=self.uri_for_key(dst),
            object_store_name=self.name,
        )

    def uri_for_key(self, key, protocol=None):
        check.str_param(key, 'key')
        protocol = check.opt_str_param(protocol, 'protocol', default='s3://')
        return protocol + self.bucket + '/' + '{key}'.format(key=key)



import math
import time

from dagster import (
    Field,
    InputDefinition,
    Nothing,
    Output,
    OutputDefinition,
    SolidDefinition,
    String,
    check,
)

from .configs import define_emr_run_job_flow_config
from .emr import EmrJobRunner
from .types import EMR_CLUSTER_DONE_STATES, EmrClusterState

_START = 'start'

# wait at most 24 hours by default for cluster job completion
_DEFAULT_CLUSTER_MAX_WAIT_TIME_SEC = 24 * 60 * 60


class EmrRunJobFlowSolidDefinition(SolidDefinition):
    def __init__(
        self,
        name,
        description=None,
        max_wait_time_sec=_DEFAULT_CLUSTER_MAX_WAIT_TIME_SEC,
        poll_interval_sec=5,
    ):
        name = check.str_param(name, 'name')

        description = check.opt_str_param(description, 'description', 'EMR create job flow solid.')

        def _compute_fn(context, _):

            job_runner = EmrJobRunner(region=context.solid_config.get('aws_region'))
            cluster_id = job_runner.run_job_flow(context.log, context.solid_config['job_config'])
            context.log.info('waiting for EMR cluster job flow completion...')

            max_iter = int(math.ceil(max_wait_time_sec / float(poll_interval_sec)))

            done = False
            curr_iter = 0
            while not done and curr_iter < max_iter:
                # This will take a while... cluster creation usually > 5 minutes
                time.sleep(poll_interval_sec)

                cluster = job_runner.describe_cluster(cluster_id)['Cluster']
                context.log.info(
                    'EMR cluster %s state: %s' % (cluster_id, cluster['Status']['State'])
                )
                done = EmrClusterState(cluster['Status']['State']) in EMR_CLUSTER_DONE_STATES
                curr_iter += 1

            yield Output(cluster_id)

        super(EmrRunJobFlowSolidDefinition, self).__init__(
            name=name,
            description=description,
            input_defs=[InputDefinition(_START, Nothing)],
            output_defs=[OutputDefinition(String)],
            compute_fn=_compute_fn,
            config={
                'aws_region': Field(String, is_required=False),
                'job_config': define_emr_run_job_flow_config(),
            },
        )



from dagster_aws.emr import EmrRunJobFlowSolidDefinition
from moto import mock_emr

from dagster import execute_pipeline, pipeline


@mock_emr
def test_run_emr_job(emr_cluster_config):
    @pipeline
    def test_pipe():
        EmrRunJobFlowSolidDefinition('test', max_wait_time_sec=2, poll_interval_sec=1)()

    config = {
        'solids': {
            'test': {'config': {'job_config': emr_cluster_config, 'aws_region': 'us-west-1'}}
        }
    }
    result = execute_pipeline(test_pipe, config)
    assert result.success



import os
import uuid

import boto3
import psycopg2
import pytest
from dagster_aws.redshift import FakeRedshiftResource, fake_redshift_resource, redshift_resource

from dagster import ModeDefinition, execute_solid, solid
from dagster.seven import mock

REDSHIFT_ENV = {
    'resources': {
        'redshift': {
            'config': {
                'host': 'foo',
                'port': 5439,
                'user': 'dagster',
                'password': 'baz',
                'database': 'dev',
                'schema': 'foobar',
            }
        }
    }
}

QUERY_RESULT = [(1,)]


def mock_execute_query_conn(*_args, **_kwargs):
    cursor_mock = mock.MagicMock(rowcount=1)
    cursor_mock.fetchall.return_value = QUERY_RESULT
    conn = mock.MagicMock(is_conn='yup')
    conn.cursor.return_value.__enter__.return_value = cursor_mock
    m = mock.MagicMock()
    m.return_value.__enter__.return_value = conn
    m.return_value = conn
    return m


@solid(required_resource_keys={'redshift'})
def single_redshift_solid(context):
    assert context.resources.redshift
    return context.resources.redshift.execute_query('SELECT 1', fetch_results=True)


@solid(required_resource_keys={'redshift'})
def multi_redshift_solid(context):
    assert context.resources.redshift
    return context.resources.redshift.execute_queries(
        ['SELECT 1', 'SELECT 1', 'SELECT 1'], fetch_results=True
    )


@mock.patch('psycopg2.connect', new_callable=mock_execute_query_conn)
def test_single_select(redshift_connect):
    result = execute_solid(
        single_redshift_solid,
        environment_dict=REDSHIFT_ENV,
        mode_def=ModeDefinition(resource_defs={'redshift': redshift_resource}),
    )
    redshift_connect.assert_called_once_with(
        host='foo',
        port=5439,
        user='dagster',
        password='baz',
        database='dev',
        schema='foobar',
        connect_timeout=5,
        sslmode='require',
    )

    assert result.success
    assert result.output_value() == QUERY_RESULT


@mock.patch('psycopg2.connect', new_callable=mock_execute_query_conn)
def test_multi_select(_redshift_connect):
    result = execute_solid(
        multi_redshift_solid,
        environment_dict=REDSHIFT_ENV,
        mode_def=ModeDefinition(resource_defs={'redshift': redshift_resource}),
    )
    assert result.success
    assert result.output_value() == [QUERY_RESULT] * 3


def test_fake_redshift():
    fake_mode = ModeDefinition(resource_defs={'redshift': fake_redshift_resource})

    result = execute_solid(single_redshift_solid, environment_dict=REDSHIFT_ENV, mode_def=fake_mode)
    assert result.success
    assert result.output_value() == FakeRedshiftResource.QUERY_RESULT

    result = execute_solid(multi_redshift_solid, environment_dict=REDSHIFT_ENV, mode_def=fake_mode)
    assert result.success
    assert result.output_value() == [FakeRedshiftResource.QUERY_RESULT] * 3


REDSHIFT_CREATE_TABLE_QUERY = '''CREATE TABLE IF NOT EXISTS VENUE1(
VENUEID SMALLINT,
VENUENAME VARCHAR(100),
VENUECITY VARCHAR(30),
VENUESTATE CHAR(2),
VENUESEATS INTEGER
) DISTSTYLE EVEN;
'''

REDSHIFT_LOAD_FILE_CONTENTS = b'''
7|BMO Field|Toronto|ON|0
16|TD Garden|Boston|MA|0
23|The Palace of Auburn Hills|Auburn Hills|MI|0
28|American Airlines Arena|Miami|FL|0
37|Staples Center|Los Angeles|CA|0
42|FedExForum|Memphis|TN|0
52|PNC Arena|Raleigh|NC  ,25   |0
59|Scotiabank Saddledome|Calgary|AB|0
66|SAP Center|San Jose|CA|0
73|Heinz Field|Pittsburgh|PA|65050
'''.strip()

REDSHIFT_FAILED_LOAD_QUERY = '''
SELECT le.query,
TRIM(le.err_reason) AS err_reason,
TRIM(le.filename) AS filename,
le.line_number AS line_number,
le.raw_line AS raw_line,
le.raw_field_value AS raw_value
FROM stl_load_errors le
WHERE le.query
AND le.query = pg_last_copy_id()
LIMIT 1;
'''


@pytest.mark.skipif(
    'AWS_REDSHIFT_TEST_DO_IT_LIVE' not in os.environ,
    reason='This test only works with a live Redshift cluster',
)
def test_live_redshift(s3_bucket):
    '''
    This test is based on:

    https://aws.amazon.com/premiumsupport/knowledge-center/redshift-stl-load-errors/

    Requires the following environment variables:

    AWS_ACCOUNT_ID - AWS account ID to use
    REDSHIFT_LOAD_IAM_ROLE - IAM role to use for Redshift load
    REDSHIFT_ENDPOINT - Redshift URL
    REDSHIFT_PASSWORD - Redshift password

    '''

    # Put file to load on S3
    file_key = uuid.uuid4().hex
    client = boto3.client('s3')
    client.put_object(Body=REDSHIFT_LOAD_FILE_CONTENTS, Bucket=s3_bucket, Key=file_key)

    @solid(required_resource_keys={'redshift'})
    def query(context):
        assert context.resources.redshift

        # First, create table:
        context.resources.redshift.execute_query(REDSHIFT_CREATE_TABLE_QUERY)

        def error_callback(error, cursor, _log):
            assert (
                str(error).strip()
                == "Load into table 'venue1' failed.  Check 'stl_load_errors' system table for details."
            )
            cursor.execute(REDSHIFT_FAILED_LOAD_QUERY)
            res = cursor.fetchall()
            assert res[0][1] == 'Char length exceeds DDL length'
            assert res[0][2] == 's3://{s3_bucket}/{file_key}'.format(
                s3_bucket=s3_bucket, file_key=file_key
            )
            assert res[0][3] == 7
            assert res[0][4].strip() == '52|PNC Arena|Raleigh|NC  ,25   |0'
            assert res[0][5].strip() == 'NC  ,25'
            raise error

        return context.resources.redshift.execute_query(
            '''COPY venue1 FROM 's3://{s3_bucket}/{file_key}'
            IAM_ROLE 'arn:aws:iam::{AWS_ACCOUNT_ID}:role/{REDSHIFT_LOAD_IAM_ROLE}'
            DELIMITER '|';
            '''.format(
                s3_bucket=s3_bucket,
                file_key=file_key,
                AWS_ACCOUNT_ID=os.environ['AWS_ACCOUNT_ID'],
                REDSHIFT_LOAD_IAM_ROLE=os.environ['REDSHIFT_LOAD_IAM_ROLE'],
            ),
            fetch_results=True,
            error_callback=error_callback,
        )

    with pytest.raises(psycopg2.InternalError):
        execute_solid(
            query,
            environment_dict={
                'resources': {
                    'redshift': {
                        'config': {
                            'host': {'env': 'REDSHIFT_ENDPOINT'},
                            'port': 5439,
                            'user': 'dagster',
                            'password': {'env': 'REDSHIFT_PASSWORD'},
                            'database': 'dev',
                        }
                    }
                }
            },
            mode_def=ModeDefinition(resource_defs={'redshift': redshift_resource}),
        )



# Portions of this file are copied from the Yelp MRJob project:
#
#   https://github.com/Yelp/mrjob
#
#
# Copyright 2009-2013 Yelp, David Marin
# Copyright 2015 Yelp
# Copyright 2017 Yelp
# Copyright 2018 Contributors
# Copyright 2019 Yelp and Contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import gzip
import re
import time
from io import BytesIO

import boto3
import six
from botocore.exceptions import WaiterError
from dagster_aws.utils.mrjob.utils import _boto3_now, _wrap_aws_client, strip_microseconds

import dagster
from dagster import check
from dagster.seven import urlparse

from .types import EMR_CLUSTER_TERMINATED_STATES, EmrClusterState, EmrStepState

# if we can't create or find our own service role, use the one
# created by the AWS console and CLI
_FALLBACK_SERVICE_ROLE = 'EMR_DefaultRole'

# if we can't create or find our own instance profile, use the one
# created by the AWS console and CLI
_FALLBACK_INSTANCE_PROFILE = 'EMR_EC2_DefaultRole'


class EmrError(Exception):
    pass


class EmrJobRunner:
    def __init__(
        self, region, check_cluster_every=30, aws_access_key_id=None, aws_secret_access_key=None,
    ):
        '''This object encapsulates various utilities for interacting with EMR clusters and invoking
        steps (jobs) on them.

        See also :py:class:`~dagster_aws.emr.EmrPySparkResource`, which wraps this job runner in a
        resource for pyspark workloads.

        Args:
            region (str): AWS region to use
            check_cluster_every (int, optional): How frequently to poll boto3 APIs for updates.
                Defaults to 30 seconds.
            aws_access_key_id ([type], optional): AWS access key ID. Defaults to None, which will
                use the default boto3 credentials chain.
            aws_secret_access_key ([type], optional): AWS secret access key. Defaults to None, which
                will use the default boto3 credentials chain.
        '''
        self.region = check.str_param(region, 'region')

        # This is in seconds
        self.check_cluster_every = check.int_param(check_cluster_every, 'check_cluster_every')
        self.aws_access_key_id = check.opt_str_param(aws_access_key_id, 'aws_access_key_id')
        self.aws_secret_access_key = check.opt_str_param(
            aws_secret_access_key, 'aws_secret_access_key'
        )

    def make_emr_client(self):
        '''Creates a boto3 EMR client. Construction is wrapped in retries in case client connection
        fails transiently.

        Returns:
            botocore.client.EMR: An EMR client
        '''
        raw_emr_client = boto3.client(
            'emr',
            aws_access_key_id=self.aws_access_key_id,
            aws_secret_access_key=self.aws_secret_access_key,
            region_name=self.region,
        )
        return _wrap_aws_client(raw_emr_client, min_backoff=self.check_cluster_every)

    def cluster_id_from_name(self, cluster_name):
        '''Get a cluster ID in the format "j-123ABC123ABC1" given a cluster name "my cool cluster".

        Args:
            cluster_name (str): The name of the cluster for which to find an ID

        Returns:
            str: The ID of the cluster

        Raises:
            EmrError: No cluster with the specified name exists
        '''
        check.str_param(cluster_name, 'cluster_name')

        response = self.make_emr_client().list_clusters().get('Clusters', [])
        for cluster in response:
            if cluster['Name'] == cluster_name:
                return cluster['Id']

        raise EmrError(
            'cluster {cluster_name} not found in region {region}'.format(
                cluster_name=cluster_name, region=self.region
            )
        )

    @staticmethod
    def construct_step_dict_for_command(step_name, command, action_on_failure='CONTINUE'):
        '''Construct an EMR step definition which uses command-runner.jar to execute a shell command
        on the EMR master.

        Args:
            step_name (str): The name of the EMR step (will show up in the EMR UI)
            command (str): The shell command to execute with command-runner.jar
            action_on_failure (str, optional): Configure action on failure (e.g., continue, or
                terminate the cluster). Defaults to 'CONTINUE'.

        Returns:
            dict: Step definition dict
        '''
        check.str_param(step_name, 'step_name')
        check.list_param(command, 'command', of_type=str)
        check.str_param(action_on_failure, 'action_on_failure')

        return {
            'Name': step_name,
            'ActionOnFailure': action_on_failure,
            'HadoopJarStep': {'Jar': 'command-runner.jar', 'Args': command},
        }

    def add_tags(self, log, tags, cluster_id):
        '''Add tags in the dict tags to cluster cluster_id.

        Args:
            log (DagsterLogManager): Log manager, for logging
            tags (dict): Dictionary of {'key': 'value'} tags
            cluster_id (str): The ID of the cluster to tag
        '''
        check.dict_param(tags, 'tags')
        check.str_param(cluster_id, 'cluster_id')

        tags_items = sorted(tags.items())

        self.make_emr_client().add_tags(
            ResourceId=cluster_id, Tags=[dict(Key=k, Value=v) for k, v in tags_items]
        )

        log.info(
            'Added EMR tags to cluster %s: %s'
            % (cluster_id, ', '.join('%s=%s' % (tag, value) for tag, value in tags_items))
        )

    def run_job_flow(self, log, cluster_config):
        '''Create an empty cluster on EMR, and return the ID of that job flow.

        Args:
            log (DagsterLogManager): Log manager, for logging
            cluster_config (dict): Configuration for this EMR job flow. See:
                https://docs.aws.amazon.com/emr/latest/APIReference/API_RunJobFlow.html

        Returns:
            str: The cluster ID, e.g. "j-ZKIY4CKQRX72"
        '''
        check.dict_param(cluster_config, 'cluster_config')

        log.debug('Creating Elastic MapReduce cluster')
        emr_client = self.make_emr_client()

        log.debug(
            'Calling run_job_flow(%s)'
            % (', '.join('%s=%r' % (k, v) for k, v in sorted(cluster_config.items())))
        )
        cluster_id = emr_client.run_job_flow(**cluster_config)['JobFlowId']

        log.info('Created new cluster %s' % cluster_id)

        # set EMR tags for the cluster
        tags = cluster_config.get('Tags', {})
        tags['__dagster_version'] = dagster.__version__
        self.add_tags(log, tags, cluster_id)
        return cluster_id

    def describe_cluster(self, cluster_id):
        '''Thin wrapper over boto3 describe_cluster.

        Args:
            cluster_id (str): Cluster to inspect

        Returns:
            dict: The cluster info. See:
                https://docs.aws.amazon.com/emr/latest/APIReference/API_DescribeCluster.html
        '''
        check.str_param(cluster_id, 'cluster_id')

        emr_client = self.make_emr_client()
        return emr_client.describe_cluster(ClusterId=cluster_id)

    def describe_step(self, cluster_id, step_id):
        '''Thin wrapper over boto3 describe_step.

        Args:
            cluster_id (str): Cluster to inspect
            step_id (str): Step ID to describe

        Returns:
            dict: The step info. See:
                https://docs.aws.amazon.com/emr/latest/APIReference/API_DescribeStep.html
        '''
        check.str_param(cluster_id, 'cluster_id')
        check.str_param(step_id, 'step_id')

        emr_client = self.make_emr_client()
        return emr_client.describe_step(ClusterId=cluster_id, StepId=step_id)

    def add_job_flow_steps(self, log, cluster_id, step_defs):
        '''Submit the constructed job flow steps to EMR for execution.

        Args:
            log (DagsterLogManager): Log manager, for logging
            cluster_id (str): The ID of the cluster
            step_defs (List[dict]): List of steps; see also `construct_step_dict_for_command`

        Returns:
            List[str]: list of step IDs.
        '''
        check.str_param(cluster_id, 'cluster_id')
        check.list_param(step_defs, 'step_defs', of_type=dict)

        emr_client = self.make_emr_client()

        steps_kwargs = dict(JobFlowId=cluster_id, Steps=step_defs)
        log.debug(
            'Calling add_job_flow_steps(%s)'
            % ','.join(('%s=%r' % (k, v)) for k, v in steps_kwargs.items())
        )
        return emr_client.add_job_flow_steps(**steps_kwargs)['StepIds']

    def wait_for_emr_steps_to_complete(self, log, cluster_id, emr_step_ids):
        '''Wait for every step of the job to complete, one by one.

        Args:
            log (DagsterLogManager): Log manager, for logging
            cluster_id (str): The ID of the cluster
            step_ids (List[str]): List of EMR step IDs to wait for
        '''
        check.str_param(cluster_id, 'cluster_id')
        check.list_param(emr_step_ids, 'step_ids', of_type=str)

        for emr_step_id in emr_step_ids:
            log.info('Waiting for EMR step %s to complete...' % emr_step_id)
            self._wait_for_emr_step_to_complete(log, cluster_id, emr_step_id)

    def _wait_for_emr_step_to_complete(self, log, cluster_id, emr_step_id):
        '''Helper for wait_for_steps_to_complete(). Wait for step with the given ID to complete.

        Args:
            cluster_id (str): The ID of the cluster
            emr_step_id (str): EMR Step ID to wait for

        Raises:
            EmrError: Raised when the step is marked by EMR as failed instead of completing
                successfully.
        '''
        check.str_param(cluster_id, 'cluster_id')
        check.str_param(emr_step_id, 'emr_step_id')

        while True:
            # don't antagonize EMR's throttling
            log.debug('Waiting %.1f seconds...' % self.check_cluster_every)
            time.sleep(self.check_cluster_every)

            step = self.describe_step(cluster_id, emr_step_id)['Step']
            step_state = EmrStepState(step['Status']['State'])

            if step_state == EmrStepState.Pending:
                cluster = self.describe_cluster(cluster_id)['Cluster']

                reason = _get_reason(cluster)
                reason_desc = (': %s' % reason) if reason else ''

                log.info('PENDING (cluster is %s%s)' % (cluster['Status']['State'], reason_desc))
                continue

            elif step_state == EmrStepState.Running:
                time_running_desc = ''

                start = step['Status']['Timeline'].get('StartDateTime')
                if start:
                    time_running_desc = ' for %s' % strip_microseconds(_boto3_now() - start)

                log.info('RUNNING%s' % time_running_desc)
                continue

            # we're done, will return at the end of this
            elif step_state == EmrStepState.Completed:
                log.info('COMPLETED')
                return
            else:
                # step has failed somehow. *reason* seems to only be set
                # when job is cancelled (e.g. 'Job terminated')
                reason = _get_reason(step)
                reason_desc = (' (%s)' % reason) if reason else ''

                log.info('%s%s' % (step_state.value, reason_desc))

                # print cluster status; this might give more context
                # why step didn't succeed
                cluster = self.describe_cluster(cluster_id)['Cluster']
                reason = _get_reason(cluster)
                reason_desc = (': %s' % reason) if reason else ''
                log.info(
                    'Cluster %s %s %s%s'
                    % (
                        cluster['Id'],
                        'was' if 'ED' in cluster['Status']['State'] else 'is',
                        cluster['Status']['State'],
                        reason_desc,
                    )
                )

                if EmrClusterState(cluster['Status']['State']) in EMR_CLUSTER_TERMINATED_STATES:
                    # was it caused by IAM roles?
                    self._check_for_missing_default_iam_roles(log, cluster)

                    # TODO: extract logs here to surface failure reason
                    # See: https://github.com/dagster-io/dagster/issues/1954

            if step_state == EmrStepState.Failed:
                log.info('EMR step %s failed' % emr_step_id)

            raise EmrError('EMR step failed')

    def _check_for_missing_default_iam_roles(self, log, cluster):
        '''If cluster couldn't start due to missing IAM roles, tell user what to do.'''

        check.dict_param(cluster, 'cluster')

        reason = _get_reason(cluster)
        if any(
            reason.endswith('/%s is invalid' % role)
            for role in (_FALLBACK_INSTANCE_PROFILE, _FALLBACK_SERVICE_ROLE)
        ):
            log.warning(
                'IAM roles are missing. See documentation for IAM roles on EMR here: '
                'https://docs.aws.amazon.com/emr/latest/ManagementGuide/emr-iam-roles.html'
            )

    def log_location_for_cluster(self, cluster_id):
        '''EMR clusters are typically launched with S3 logging configured. This method inspects a
        cluster using boto3 describe_cluster to retrieve the log URI.

        Args:
            cluster_id (str): The cluster to inspect.

        Raises:
            EmrError: the log URI was missing (S3 log mirroring not enabled for this cluster)

        Returns:
            (str, str): log bucket and key
        '''
        check.str_param(cluster_id, 'cluster_id')

        # The S3 log URI is specified per job flow (cluster)
        log_uri = self.describe_cluster(cluster_id)['Cluster'].get('LogUri', None)

        # ugh, seriously boto3?! This will come back as string "None"
        if log_uri == 'None' or log_uri is None:
            raise EmrError('Log URI not specified, cannot retrieve step execution logs')

        # For some reason the API returns an s3n:// protocol log URI instead of s3://
        log_uri = re.sub('^s3n', 's3', log_uri)
        log_uri_parsed = urlparse(log_uri)
        log_bucket = log_uri_parsed.netloc
        log_key_prefix = log_uri_parsed.path.lstrip('/')
        return log_bucket, log_key_prefix

    def retrieve_logs_for_step_id(self, log, cluster_id, step_id):
        '''Retrieves stdout and stderr logs for the given step ID.

        Args:
            log (DagsterLogManager): Log manager, for logging
            cluster_id (str): EMR cluster ID
            step_id (str): EMR step ID for the job that was submitted.

        Returns
            (str, str): Tuple of stdout log string contents, and stderr log string contents
        '''
        check.str_param(cluster_id, 'cluster_id')
        check.str_param(step_id, 'step_id')

        log_bucket, log_key_prefix = self.log_location_for_cluster(cluster_id)

        prefix = '{log_key_prefix}{cluster_id}/steps/{step_id}'.format(
            log_key_prefix=log_key_prefix, cluster_id=cluster_id, step_id=step_id
        )
        stdout_log = self.wait_for_log(log, log_bucket, '{prefix}/stdout.gz'.format(prefix=prefix))
        stderr_log = self.wait_for_log(log, log_bucket, '{prefix}/stderr.gz'.format(prefix=prefix))
        return stdout_log, stderr_log

    def wait_for_log(self, log, log_bucket, log_key, waiter_delay=30, waiter_max_attempts=20):
        '''Wait for gzipped EMR logs to appear on S3. Note that EMR syncs logs to S3 every 5
        minutes, so this may take a long time.

        Args:
            log_bucket (str): S3 bucket where log is expected to appear
            log_key (str): S3 key for the log file
            waiter_delay (int): How long to wait between attempts to check S3 for the log file
            waiter_max_attempts (int): Number of attempts before giving up on waiting

        Raises:
            EmrError: Raised if we waited the full duration and the logs did not appear

        Returns:
            str: contents of the log file
        '''
        check.str_param(log_bucket, 'log_bucket')
        check.str_param(log_key, 'log_key')
        check.int_param(waiter_delay, 'waiter_delay')
        check.int_param(waiter_max_attempts, 'waiter_max_attempts')

        log.info(
            'Attempting to get log: s3://{log_bucket}/{log_key}'.format(
                log_bucket=log_bucket, log_key=log_key
            )
        )

        s3 = _wrap_aws_client(boto3.client('s3'), min_backoff=self.check_cluster_every)
        waiter = s3.get_waiter('object_exists')
        try:
            waiter.wait(
                Bucket=log_bucket,
                Key=log_key,
                WaiterConfig={'Delay': waiter_delay, 'MaxAttempts': waiter_max_attempts},
            )
        except WaiterError as err:
            six.raise_from(
                EmrError('EMR log file did not appear on S3 after waiting'), err,
            )
        obj = BytesIO(s3.get_object(Bucket=log_bucket, Key=log_key)['Body'].read())
        gzip_file = gzip.GzipFile(fileobj=obj)
        return gzip_file.read().decode('utf-8')


def _get_reason(cluster_or_step):
    '''Get state change reason message.'''
    # StateChangeReason is {} before the first state change
    return cluster_or_step['Status']['StateChangeReason'].get('Message', '')



import pytest


@pytest.fixture(scope='session')
def emr_cluster_config():
    return {
        'Name': 'test-emr',
        'LogUri': 's3n://emr-cluster-logs/elasticmapreduce/',
        'ReleaseLabel': 'emr-5.23.0',
        'Instances': {
            'MasterInstanceType': 'm4.large',
            'SlaveInstanceType': 'm4.large',
            'InstanceCount': 4,
            'TerminationProtected': False,
        },
        'Applications': [{'Name': 'Spark'}],
        'BootstrapActions': [
            {
                'Name': 'Spark Default Config',
                'ScriptBootstrapAction': {
                    'Path': 's3://support.elasticmapreduce/spark/maximize-spark-default-config'
                },
            }
        ],
        'Steps': [
            {
                'Name': 'Setup Debugging',
                'ActionOnFailure': 'TERMINATE_CLUSTER',
                'HadoopJarStep': {'Jar': 'command-runner.jar', 'Args': ['state-pusher-script']},
            },
        ],
        'VisibleToAllUsers': True,
        'JobFlowRole': 'EMR_EC2_DefaultRole',
        'ServiceRole': 'EMR_DefaultRole',
    }



import boto3
import pytest

TEST_CLOUDWATCH_LOG_GROUP_NAME = '/dagster-test/test-cloudwatch-logging'
TEST_CLOUDWATCH_LOG_STREAM_NAME = 'test-logging'
AWS_REGION = 'us-west-1'


@pytest.fixture(scope='session')
def cloudwatch_client():
    client = boto3.client('logs', AWS_REGION)
    client.describe_log_groups(logGroupNamePrefix=TEST_CLOUDWATCH_LOG_GROUP_NAME)
    res = client.describe_log_streams(
        logGroupName=TEST_CLOUDWATCH_LOG_GROUP_NAME,
        logStreamNamePrefix=TEST_CLOUDWATCH_LOG_STREAM_NAME,
    )
    assert TEST_CLOUDWATCH_LOG_STREAM_NAME in (
        log_stream['logStreamName'] for log_stream in res['logStreams']
    )
    yield client



import copy
import gzip
import io
import threading
import time

import boto3
import pytest
from dagster_aws.emr import EmrClusterState, EmrError, EmrJobRunner
from dagster_aws.utils.mrjob.utils import _boto3_now
from moto import mock_emr, mock_s3

from dagster.seven import mock
from dagster.utils.test import create_test_pipeline_execution_context

REGION = 'us-west-1'


@mock_emr
def test_emr_create_cluster(emr_cluster_config):
    context = create_test_pipeline_execution_context()
    cluster = EmrJobRunner(region=REGION)
    cluster_id = cluster.run_job_flow(context.log, emr_cluster_config)
    assert cluster_id.startswith('j-')


@mock_emr
def test_emr_add_tags_and_describe_cluster(emr_cluster_config):
    context = create_test_pipeline_execution_context()
    emr = EmrJobRunner(region=REGION)

    cluster_id = emr.run_job_flow(context.log, emr_cluster_config)

    emr.add_tags(context.log, {'foobar': 'v1', 'baz': '123'}, cluster_id)

    tags = emr.describe_cluster(cluster_id)['Cluster']['Tags']

    assert {'Key': 'baz', 'Value': '123'} in tags
    assert {'Key': 'foobar', 'Value': 'v1'} in tags


@mock_emr
def test_emr_describe_cluster(emr_cluster_config):
    context = create_test_pipeline_execution_context()
    cluster = EmrJobRunner(region=REGION)
    cluster_id = cluster.run_job_flow(context.log, emr_cluster_config)
    cluster_info = cluster.describe_cluster(cluster_id)['Cluster']
    assert cluster_info['Name'] == 'test-emr'
    assert EmrClusterState(cluster_info['Status']['State']) == EmrClusterState.Waiting


@mock_emr
def test_emr_id_from_name(emr_cluster_config):
    context = create_test_pipeline_execution_context()
    cluster = EmrJobRunner(region=REGION)
    cluster_id = cluster.run_job_flow(context.log, emr_cluster_config)
    assert cluster.cluster_id_from_name('test-emr') == cluster_id

    with pytest.raises(EmrError) as exc_info:
        cluster.cluster_id_from_name('cluster-doesnt-exist')

    assert 'cluster cluster-doesnt-exist not found in region us-west-1' in str(exc_info.value)


def test_emr_construct_step_dict():
    cmd = ['pip', 'install', 'dagster']

    assert EmrJobRunner.construct_step_dict_for_command('test_step', cmd) == {
        'Name': 'test_step',
        'ActionOnFailure': 'CONTINUE',
        'HadoopJarStep': {'Jar': 'command-runner.jar', 'Args': cmd},
    }

    assert EmrJobRunner.construct_step_dict_for_command(
        'test_second_step', cmd, action_on_failure='CANCEL_AND_WAIT'
    ) == {
        'Name': 'test_second_step',
        'ActionOnFailure': 'CANCEL_AND_WAIT',
        'HadoopJarStep': {'Jar': 'command-runner.jar', 'Args': cmd},
    }


@mock_emr
def test_emr_log_location_for_cluster(emr_cluster_config):
    context = create_test_pipeline_execution_context()
    emr = EmrJobRunner(region=REGION)
    cluster_id = emr.run_job_flow(context.log, emr_cluster_config)
    assert emr.log_location_for_cluster(cluster_id) == ('emr-cluster-logs', 'elasticmapreduce/')

    # Should raise when the log URI is missing
    emr_cluster_config = copy.deepcopy(emr_cluster_config)
    del emr_cluster_config['LogUri']
    cluster_id = emr.run_job_flow(context.log, emr_cluster_config)
    with pytest.raises(EmrError) as exc_info:
        emr.log_location_for_cluster(cluster_id)

    assert 'Log URI not specified, cannot retrieve step execution logs' in str(exc_info.value)


@mock_emr
@mock_s3
def test_emr_retrieve_logs(emr_cluster_config):
    context = create_test_pipeline_execution_context()
    emr = EmrJobRunner(region=REGION)
    cluster_id = emr.run_job_flow(context.log, emr_cluster_config)
    assert emr.log_location_for_cluster(cluster_id) == ('emr-cluster-logs', 'elasticmapreduce/')

    s3 = boto3.resource('s3', region_name=REGION)
    s3.create_bucket(Bucket='emr-cluster-logs')  # pylint: disable=no-member

    def create_log():
        time.sleep(0.5)
        out = io.BytesIO()
        with gzip.GzipFile(fileobj=out, mode='w') as fo:
            fo.write('some log'.encode())

        prefix = 'elasticmapreduce/{cluster_id}/steps/{step_id}'.format(
            cluster_id=cluster_id, step_id='s-123456123456'
        )

        for name in ['stdout.gz', 'stderr.gz']:
            s3.Object('emr-cluster-logs', prefix + '/' + name).put(  # pylint: disable=no-member
                Body=out.getvalue()
            )

    thread = threading.Thread(target=create_log, args=())
    thread.daemon = True
    thread.start()

    stdout_log, stderr_log = emr.retrieve_logs_for_step_id(
        context.log, cluster_id, 's-123456123456'
    )
    assert stdout_log == 'some log'
    assert stderr_log == 'some log'


@mock_s3
def test_wait_for_log():
    s3 = boto3.resource('s3', region_name=REGION)
    s3.create_bucket(Bucket='log_bucket')  # pylint: disable=no-member

    def create_log():
        time.sleep(0.5)
        out = io.BytesIO()
        with gzip.GzipFile(fileobj=out, mode='w') as fo:
            fo.write('foo bar'.encode())

        s3.Object('log_bucket', 'some_log_file').put(  # pylint: disable=no-member
            Body=out.getvalue()
        )

    thread = threading.Thread(target=create_log, args=())
    thread.daemon = True
    thread.start()

    context = create_test_pipeline_execution_context()
    emr = EmrJobRunner(region=REGION)
    res = emr.wait_for_log(
        context.log,
        log_bucket='log_bucket',
        log_key='some_log_file',
        waiter_delay=1,
        waiter_max_attempts=2,
    )
    assert res == 'foo bar'

    with pytest.raises(EmrError) as exc_info:
        emr.wait_for_log(
            context.log,
            log_bucket='log_bucket',
            log_key='does_not_exist',
            waiter_delay=1,
            waiter_max_attempts=1,
        )
    assert 'EMR log file did not appear on S3 after waiting' in str(exc_info.value)


@mock_emr
def test_emr_wait_for_step(emr_cluster_config):
    context = create_test_pipeline_execution_context()
    emr = EmrJobRunner(region=REGION, check_cluster_every=1)

    cluster_id = emr.run_job_flow(context.log, emr_cluster_config)

    step_name = 'test_step'
    step_cmd = ['ls', '/']
    step_ids = emr.add_job_flow_steps(
        context.log, cluster_id, [emr.construct_step_dict_for_command(step_name, step_cmd)]
    )

    def get_step_dict(step_id, step_state):
        return {
            'Step': {
                'Id': step_id,
                'Name': step_name,
                'Config': {'Jar': 'command-runner.jar', 'Properties': {}, 'Args': step_cmd},
                'ActionOnFailure': 'CONTINUE',
                'Status': {
                    'State': step_state,
                    'StateChangeReason': {'Message': 'everything is hosed'},
                    'Timeline': {'StartDateTime': _boto3_now()},
                },
            },
        }

    calls = {'num_calls': 0, 'final_state': 'COMPLETED'}

    def new_describe_step(_, cluster_id, step_id):
        calls['num_calls'] += 1

        if calls['num_calls'] == 1:
            return get_step_dict(step_id, 'PENDING')
        elif calls['num_calls'] == 2:
            return get_step_dict(step_id, 'RUNNING')
        else:
            return get_step_dict(step_id, calls['final_state'])

        return emr.describe_step(cluster_id, step_id)

    with mock.patch.object(EmrJobRunner, 'describe_step', new=new_describe_step):
        emr.wait_for_emr_steps_to_complete(context.log, cluster_id, step_ids)

    calls['num_calls'] = 0
    calls['final_state'] = 'FAILED'
    with pytest.raises(EmrError) as exc_info:
        with mock.patch.object(EmrJobRunner, 'describe_step', new=new_describe_step):
            emr.wait_for_emr_steps_to_complete(context.log, cluster_id, step_ids)
    assert 'step failed' in str(exc_info.value)



from dagster_aws.utils.mrjob.log4j import Log4jRecord, parse_hadoop_log4j_records

SIMPLE_LOG4J = '''
20/02/05 17:26:43 INFO SparkContext: Running Spark version 2.4.4
20/02/05 17:26:43 INFO SparkContext: Submitted application: blah
20/02/05 17:26:43 INFO SecurityManager: Changing view acls to: hadoop
'''.strip()

MULTILINE_LOG4J = '''
20/02/05 17:26:50 INFO Client: Application report for application_1580918830280_0002 (state: ACCEPTED)
20/02/05 17:26:50 INFO Client:
         client token: N/A
         diagnostics: AM container is launched, waiting for AM container to Register with RM
         ApplicationMaster host: N/A
         ApplicationMaster RPC port: -1
         queue: default
         start time: 1580923609467
         final status: UNDEFINED
         tracking URL: http://ip-172-31-2-74.us-west-1.compute.internal:20888/proxy/application_1580918830280_0002/
         user: hadoop
20/02/05 17:26:51 INFO Client: Application report for application_1580918830280_0002 (state: ACCEPTED)
'''.strip()


def test_simple_log4j_parsing():
    res = parse_hadoop_log4j_records(SIMPLE_LOG4J)
    expected = [
        Log4jRecord(
            caller_location='',
            level='INFO',
            logger='SparkContext',
            message='Running Spark version 2.4.4',
            num_lines=1,
            start_line=0,
            thread=None,
            timestamp='20/02/05 17:26:43',
        ),
        Log4jRecord(
            caller_location='',
            level='INFO',
            logger='SparkContext',
            message='Submitted application: blah',
            num_lines=1,
            start_line=1,
            thread=None,
            timestamp='20/02/05 17:26:43',
        ),
        Log4jRecord(
            caller_location='',
            level='INFO',
            logger='SecurityManager',
            message='Changing view acls to: hadoop',
            num_lines=1,
            start_line=2,
            thread=None,
            timestamp='20/02/05 17:26:43',
        ),
    ]
    assert list(res) == expected


def test_multiline_log4j_parsing():
    res = parse_hadoop_log4j_records(MULTILINE_LOG4J)

    expected = [
        Log4jRecord(
            caller_location='',
            level='INFO',
            logger='Client',
            message='Application report for application_1580918830280_0002 (state: ACCEPTED)',
            num_lines=1,
            start_line=0,
            thread=None,
            timestamp='20/02/05 17:26:50',
        ),
        Log4jRecord(
            caller_location='',
            level='INFO',
            logger='Client',
            message='\n'.join(
                [
                    '',
                    '         client token: N/A',
                    '         diagnostics: AM container is launched, waiting for AM container to Register with RM',
                    '         ApplicationMaster host: N/A',
                    '         ApplicationMaster RPC port: -1',
                    '         queue: default',
                    '         start time: 1580923609467',
                    '         final status: UNDEFINED',
                    '         tracking URL: http://ip-172-31-2-74.us-west-1.compute.internal:20888/proxy/application_1580918830280_0002/',
                    '         user: hadoop',
                ]
            ),
            num_lines=10,
            start_line=1,
            thread=None,
            timestamp='20/02/05 17:26:50',
        ),
        Log4jRecord(
            caller_location='',
            level='INFO',
            logger='Client',
            message='Application report for application_1580918830280_0002 (state: ACCEPTED)',
            num_lines=1,
            start_line=11,
            thread=None,
            timestamp='20/02/05 17:26:51',
        ),
    ]
    assert list(res) == expected



import io

from dagster_aws.s3 import S3FakeSession, S3FileCache, S3FileHandle


def test_s3_file_cache_file_not_present():
    session_fake = S3FakeSession()
    file_store = S3FileCache(
        s3_bucket='some-bucket', s3_key='some-key', s3_session=session_fake, overwrite=False
    )

    assert not file_store.has_file_object('foo')


def test_s3_file_cache_file_present():
    session_fake = S3FakeSession()
    file_store = S3FileCache(
        s3_bucket='some-bucket', s3_key='some-key', s3_session=session_fake, overwrite=False
    )

    assert not file_store.has_file_object('foo')

    file_store.write_binary_data('foo', 'bar'.encode())

    assert file_store.has_file_object('foo')


def test_s3_file_cache_correct_handle():
    session_fake = S3FakeSession()
    file_store = S3FileCache(
        s3_bucket='some-bucket', s3_key='some-key', s3_session=session_fake, overwrite=False
    )

    assert isinstance(file_store.get_file_handle('foo'), S3FileHandle)


def test_s3_file_cache_write_file_object():
    session_fake = S3FakeSession()
    file_store = S3FileCache(
        s3_bucket='some-bucket', s3_key='some-key', s3_session=session_fake, overwrite=False
    )

    stream = io.BytesIO('content'.encode())
    file_store.write_file_object('foo', stream)



import os
from collections import namedtuple

import six
import terminaltables
import yaml

from dagster import check

from .term import Term

HOST_CONFIG_FILE = 'dagster-aws-config.yaml'


class ConfigMixin(object):
    @classmethod
    def exists(cls, base_path):
        '''Check that the configuration file exists and the key for this config type exists in the
        file
        '''
        cfg_path = os.path.join(base_path, HOST_CONFIG_FILE)
        if not os.path.exists(cfg_path):
            return False

        with open(cfg_path, 'rb') as f:
            record = yaml.safe_load(f)
            if record and cls.KEY in record:
                return True

        return False

    def save(self, base_path):
        '''Serialize configuration to a YAML file for future use
        '''
        if not os.path.exists(base_path):
            os.makedirs(base_path)

        cfg_path = os.path.join(base_path, HOST_CONFIG_FILE)

        record = {}

        if os.path.exists(cfg_path):
            with open(cfg_path, 'rb') as f:
                record = yaml.safe_load(f) or {}

        record[self.KEY] = dict(self._asdict())

        with open(cfg_path, 'wb') as f:
            f.write(six.ensure_binary(yaml.dump(record, default_flow_style=False)))
        Term.info('Saved %s configuration to %s' % (self.TITLE, cfg_path))

    @classmethod
    def load(cls, base_path):
        '''Deserialize the configuration from the YAML config file.
        '''
        filepath = os.path.join(base_path, HOST_CONFIG_FILE)

        if not os.path.exists(filepath):
            Term.fatal('No configuration found, run `dagster-aws init` to get started.')

        with open(filepath, 'rb') as f:
            raw_cfg = yaml.safe_load(f)
        return cls.__new__(cls, **raw_cfg.get(cls.KEY, {}))

    def as_table(self):
        '''Returns a tabulated string representation of this config class
        '''
        as_dict = self._asdict()
        if 'password' in as_dict:
            as_dict['password'] = '<redacted>'

        table_data = [('Config Key', 'Value')] + list(as_dict.items())
        return terminaltables.SingleTable(table_data, title=self.TITLE).table

    def delete(self, base_path):
        '''Remove the configuration for this resource from HOST_CONFIG_FILE
        '''
        if not self.exists(base_path):
            Term.warning('No configuration for %s found, skipping deletion' % self.KEY)
            return False

        cfg_path = os.path.join(base_path, HOST_CONFIG_FILE)
        with open(cfg_path, 'rb') as f:
            record = yaml.safe_load(f) or {}

        if self.KEY in record:
            del record[self.KEY]
            with open(cfg_path, 'wb') as f:
                f.write(six.ensure_binary(yaml.dump(record, default_flow_style=False)))
            Term.info('Removed configuration for %s from %s' % (self.KEY, cfg_path))
            return True

        return False


class RDSConfig(
    namedtuple(
        '_RDSConfig',
        'instance_name instance_type storage_size_gb db_engine db_engine_version instance_uri '
        'db_name username password',
    ),
    ConfigMixin,
):
    TITLE = 'RDS Configuration'
    KEY = 'rds'

    def __new__(
        cls,
        instance_name=None,
        instance_type=None,
        storage_size_gb=20,
        db_engine='postgres',
        db_engine_version='11.6',
        instance_uri=None,
        db_name='dagster',
        username='dagster',
        password=None,
    ):
        return super(RDSConfig, cls).__new__(
            cls,
            instance_name=check.opt_str_param(instance_name, 'instance_name'),
            instance_type=check.opt_str_param(instance_type, 'instance_type'),
            storage_size_gb=check.opt_int_param(storage_size_gb, 'storage_size_gb'),
            db_engine=check.opt_str_param(db_engine, 'db_engine'),
            db_engine_version=check.opt_str_param(db_engine_version, 'db_engine_version'),
            instance_uri=check.opt_str_param(instance_uri, 'instance_uri'),
            db_name=check.opt_str_param(db_name, 'db_name'),
            username=check.opt_str_param(username, 'username'),
            password=check.opt_str_param(password, 'password'),
        )


class EC2Config(
    namedtuple(
        '_HostConfig',
        'remote_host instance_id region security_group_id key_pair_name key_file_path ami_id '
        'local_path',
    ),
    ConfigMixin,
):
    '''Serialize the user's AWS host configuration to a YAML file for future use.
    '''

    TITLE = 'EC2 Configuration'
    KEY = 'ec2'

    def __new__(
        cls,
        remote_host=None,
        instance_id=None,
        region=None,
        security_group_id=None,
        key_pair_name=None,
        key_file_path=None,
        ami_id=None,
        local_path=None,
    ):
        return super(EC2Config, cls).__new__(
            cls,
            remote_host=check.opt_str_param(remote_host, 'remote_host'),
            instance_id=check.opt_str_param(instance_id, 'instance_id'),
            region=check.opt_str_param(region, 'region'),
            security_group_id=check.opt_str_param(security_group_id, 'security_group_id'),
            key_pair_name=check.opt_str_param(key_pair_name, 'key_pair_name'),
            key_file_path=check.opt_str_param(key_file_path, 'key_file_path'),
            ami_id=check.opt_str_param(ami_id, 'ami_id'),
            local_path=check.opt_str_param(local_path, 'local_path'),
        )



from dagster import Field, resource

from .utils import create_s3_session


@resource(
    {
        'use_unsigned_session': Field(
            bool,
            description='Specifies whether to use an unsigned S3 session',
            is_required=False,
            default_value=False,
        ),
        'region_name': Field(
            str, description='Specifies a custom region for the S3 session', is_required=False
        ),
        'endpoint_url': Field(
            str, description='Specifies a custom endpoint for the S3 session', is_required=False
        ),
    }
)
def s3_resource(context):
    '''Resource that gives solids access to S3.

    The underlying S3 session is created by calling :py:func:`boto3.resource('s3') <boto3:boto3.resource>`.

    Attach this resource definition to a :py:class:`~dagster.ModeDefinition` in order to make it
    available to your solids.

    Example:

        .. code-block:: python

            from dagster import ModeDefinition, execute_solid, solid
            from dagster_aws.s3 import s3_resource

            @solid(required_resource_keys={'s3'})
            def example_s3_solid(context):
                return context.resources.s3.list_objects_v2(
                    Bucket='my-bucket',
                    Prefix='some-key'
                )

            result = execute_solid(
                example_s3_solid,
                environment_dict={
                    'resources': {
                        's3': {
                            'config': {
                                'region_name': 'us-west-1',
                            }
                        }
                    }
                },
                mode_def=ModeDefinition(resource_defs={'s3': s3_resource}),
            )

    Note that your solids must also declare that they require this resource with
    `required_resource_keys`, or it will not be initialized for the execution of their compute
    functions.

    You may configure this resource as follows:

    .. code-block:: YAML

        resources:
          s3:
            config:
              region_name: "us-west-1"
              # Optional[str]: Specifies a custom region for the S3 session. Default is chosen
              # through the ordinary boto credential chain.
              use_unsigned_session: false
              # Optional[bool]: Specifies whether to use an unsigned S3 session. Default: True
              endpoint_url: "http://localhost"
              # Optional[str]: Specifies a custom endpoint for the S3 session. Default is None.
    '''
    use_unsigned_session = context.resource_config['use_unsigned_session']
    region_name = context.resource_config.get('region_name')
    endpoint_url = context.resource_config.get('endpoint_url')

    return create_s3_session(
        signed=not use_unsigned_session, region_name=region_name, endpoint_url=endpoint_url
    )
